@@ -1,10 +1,10 @@
 /*******************************************************************************
  *
  * NRPE.C - Nagios Remote Plugin Executor
- * Copyright (c) 1999-2003 Ethan Galstad (nagios@nagios.org)
+ * Copyright (c) 1999-2007 Ethan Galstad (nagios@nagios.org)
  * License: GPL
  *
- * Last Modified: 09-08-2003
+ * Last Modified: 05-10-2007
  *
  * Command line: nrpe -c <config_file> [--inetd | --daemon]
  *
@@ -18,61 +18,19 @@
  * 
  ******************************************************************************/
 
-#include "../common/common.h"
-#include "../common/config.h"
-#include "nrpe.h"
-#include "utils.h"
+#include "../include/common.h"
+#include "../include/config.h"
+#include "../include/nrpe.h"
+#include "../include/utils.h"
 
 #ifdef HAVE_SSL
-#include "dh.h"
+#include "../include/dh.h"
 #endif
 
-#define DEFAULT_COMMAND_TIMEOUT	60			/* default timeout for execution of plugins */
-#define MAXFD                   64
-#define NASTY_METACHARS         "|`&><'\"\\[]{}"
-
-
-int process_arguments(int,char **);
-void wait_for_connections(void);
-void handle_connection(int);
-int read_config_file(char *);
-int read_config_dir(char *);
-int add_command(char *,char *);
-command *find_command(char *);
-void sighandler(int);
-int drop_privileges(char *,char *);
-int check_privileges(void);
-void free_memory(void);
-int is_an_allowed_host(char *);
-int validate_request(packet *);
-int contains_nasty_metachars(char *);
-int process_macros(char *,char *,int);
-int my_system(char *,int,int *,char *,int);            	/* executes a command via popen(), but also protects against timeouts */
-void my_system_sighandler(int);				/* handles timeouts when executing commands via my_system() */
-
-
-char    *command_name=NULL;
-char    *macro_argv[MAX_COMMAND_ARGUMENTS];
-
-char    config_file[MAX_INPUT_BUFFER]="nrpe.cfg";
-char    allowed_hosts[MAX_INPUT_BUFFER];
-int     server_port=DEFAULT_SERVER_PORT;
-char    server_address[16]="0.0.0.0";
-int     socket_timeout=DEFAULT_SOCKET_TIMEOUT;
-int     command_timeout=DEFAULT_COMMAND_TIMEOUT;
-
-command *command_list=NULL;
-
-char    *nrpe_user=NULL;
-char    *nrpe_group=NULL;
-
-int     allow_arguments=FALSE;
-
-int     show_help=FALSE;
-int     show_license=FALSE;
-int     show_version=FALSE;
-int     use_inetd=TRUE;
-int     debug=FALSE;
+#ifdef HAVE_LIBWRAP
+int allow_severity=LOG_INFO;
+int deny_severity=LOG_WARNING;
+#endif
 
 #ifdef HAVE_SSL
 SSL_METHOD *meth;
@@ -82,6 +40,45 @@ int use_ssl=TRUE;
 int use_ssl=FALSE;
 #endif
 
+#define DEFAULT_COMMAND_TIMEOUT	60			/* default timeout for execution of plugins */
+#define MAXFD                   64
+#define NASTY_METACHARS         "|`&><'\"\\[]{};"
+
+char    *command_name=NULL;
+char    *macro_argv[MAX_COMMAND_ARGUMENTS];
+
+char    config_file[MAX_INPUT_BUFFER]="nrpe.cfg";
+int     server_port=DEFAULT_SERVER_PORT;
+char    server_address[16]="0.0.0.0";
+int     socket_timeout=DEFAULT_SOCKET_TIMEOUT;
+int     command_timeout=DEFAULT_COMMAND_TIMEOUT;
+int     connection_timeout=DEFAULT_CONNECTION_TIMEOUT;
+char    *command_prefix=NULL;
+
+command *command_list=NULL;
+
+char    *nrpe_user=NULL;
+char    *nrpe_group=NULL;
+
+char    *allowed_hosts=NULL;
+
+char    *pid_file=NULL;
+int     wrote_pid_file=FALSE;
+
+int     allow_arguments=FALSE;
+
+int     allow_weak_random_seed=FALSE;
+
+int     sigrestart=FALSE;
+int     sigshutdown=FALSE;
+
+int     show_help=FALSE;
+int     show_license=FALSE;
+int     show_version=FALSE;
+int     use_inetd=TRUE;
+int     debug=FALSE;
+
+
 
 
 int main(int argc, char **argv){
@@ -90,6 +87,8 @@ int main(int argc, char **argv){
 	char buffer[MAX_INPUT_BUFFER];
 #ifdef HAVE_SSL
 	DH *dh;
+	char seedfile[FILENAME_MAX];
+	int i,c;
 #endif
 
 	result=process_arguments(argc,argv);
@@ -98,12 +97,15 @@ int main(int argc, char **argv){
 
 		printf("\n");
 		printf("NRPE - Nagios Remote Plugin Executor\n");
-		printf("Copyright (c) 1999-2003 Ethan Galstad (nagios@nagios.org)\n");
+		printf("Copyright (c) 1999-2007 Ethan Galstad (nagios@nagios.org)\n");
 		printf("Version: %s\n",PROGRAM_VERSION);
 		printf("Last Modified: %s\n",MODIFICATION_DATE);
-		printf("License: GPL with exemptions (-l for more info)\n");
+		printf("License: GPL v2 with exemptions (-l for more info)\n");
 #ifdef HAVE_SSL
 		printf("SSL/TLS Available: Anonymous DH Mode, OpenSSL 0.9.6 or higher required\n");
+#endif
+#ifdef HAVE_LIBWRAP
+		printf("TCP Wrappers Available\n");
 #endif
 		printf("\n");
 #ifdef ENABLE_COMMAND_ARGUMENTS
@@ -112,6 +114,13 @@ int main(int argc, char **argv){
 		printf("**      Read the NRPE SECURITY file for more information     **\n");
 		printf("***************************************************************\n");
 		printf("\n");
+#endif
+#ifndef HAVE_LIBWRAP
+		printf("***************************************************************\n");
+		printf("** POSSIBLE SECURITY RISK - TCP WRAPPERS ARE NOT AVAILABLE!  **\n");
+		printf("**      Read the NRPE SECURITY file for more information     **\n");
+		printf("***************************************************************\n");
+		printf("\n");
 #endif
 	        }
 
@@ -120,9 +129,10 @@ int main(int argc, char **argv){
 
 	else if(result!=OK || show_help==TRUE){
 
-		printf("Usage: nrpe -c <config_file> <mode>\n");
+		printf("Usage: nrpe [-n] -c <config_file> <mode>\n");
 		printf("\n");
 		printf("Options:\n");
+		printf(" -n            = Do not use SSL\n");
 		printf(" <config_file> = Name of config file to use\n");
 		printf(" <mode>        = One of the following two operating modes:\n");  
 		printf("   -i          =    Run as a service under inetd or xinetd\n");
@@ -171,17 +181,17 @@ int main(int argc, char **argv){
 
 	/* exit if there are errors... */
 	if(result==ERROR){
-		syslog(LOG_ERR,"Config file '%s' contained errors, bailing out...",config_file);
+		syslog(LOG_ERR,"Config file '%s' contained errors, aborting...",config_file);
 		return STATE_CRITICAL;
 		}
 
+        /* generate the CRC 32 table */
+        generate_crc32_table();
+
 	/* initialize macros */
 	for(x=0;x<MAX_COMMAND_ARGUMENTS;x++)
 		macro_argv[x]=NULL;
 
-        /* generate the CRC 32 table */
-        generate_crc32_table();
-
 #ifdef HAVE_SSL
 	/* initialize SSL */
 	if(use_ssl==TRUE){
@@ -189,11 +199,36 @@ int main(int argc, char **argv){
 		SSLeay_add_ssl_algorithms();
 		meth=SSLv23_server_method();
 		SSL_load_error_strings();
+
+		/* use week random seed if necessary */
+		if(allow_weak_random_seed && (RAND_status()==0)){
+
+			if(RAND_file_name(seedfile,sizeof(seedfile)-1))
+				if(RAND_load_file(seedfile,-1))
+					RAND_write_file(seedfile);
+
+			if(RAND_status()==0){
+				syslog(LOG_ERR,"Warning: SSL/TLS uses a weak random seed which is highly discouraged");
+				srand(time(NULL));
+				for(i=0;i<500 && RAND_status()==0;i++){
+					for(c=0;c<sizeof(seedfile);c+=sizeof(int)){
+						*((int *)(seedfile+c))=rand();
+					        }
+					RAND_seed(seedfile,sizeof(seedfile));
+					}
+				}
+			}
+
 		if((ctx=SSL_CTX_new(meth))==NULL){
 			syslog(LOG_ERR,"Error: could not create SSL context.\n");
 			exit(STATE_CRITICAL);
 		        }
-		/*SSL_CTX_set_cipher_list(ctx,"ALL");*/
+
+		/* ADDED 01/19/2004 */
+		/* use only TLSv1 protocol */
+		SSL_CTX_set_options(ctx,SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
+
+		/* use anonymous DH ciphers */
 		SSL_CTX_set_cipher_list(ctx,"ADH");
 		dh=get_dh512();
 		SSL_CTX_set_tmp_dh(ctx,dh);
@@ -213,22 +248,20 @@ int main(int argc, char **argv){
 		/* make sure we're not root */
 		check_privileges();
 
+		/* redirect STDERR to /dev/null */
+		close(2);
+		open("/dev/null",O_WRONLY);
+
 		/* handle the connection */
 		handle_connection(0);
 	        }
 
 	/* else daemonize and start listening for requests... */
 	else if(fork()==0){
-
+		
 		/* we're a daemon - set up a new process group */
 		setsid();
 
-		/* ignore SIGHUP */
-		signal(SIGHUP, SIG_IGN);
-
-		chdir("/");
-		umask(0);
-
 		/* close standard file descriptors */
 		close(0);
 		close(1);
@@ -239,14 +272,57 @@ int main(int argc, char **argv){
 		open("/dev/null",O_WRONLY);
 		open("/dev/null",O_WRONLY);
 
+		chdir("/");
+		/*umask(0);*/
+
+		/* handle signals */
+		signal(SIGQUIT,sighandler);
+		signal(SIGTERM,sighandler);
+		signal(SIGHUP,sighandler);
+
+		/* log info to syslog facility */
+		syslog(LOG_NOTICE,"Starting up daemon");
+
+		/* write pid file */
+		if(write_pid_file()==ERROR)
+			return STATE_CRITICAL;
+		
 		/* drop privileges */
 		drop_privileges(nrpe_user,nrpe_group);
 
 		/* make sure we're not root */
 		check_privileges();
 
-		/* wait for connections */
-		wait_for_connections();
+		do{
+
+			/* reset flags */
+			sigrestart=FALSE;
+			sigshutdown=FALSE;
+
+			/* wait for connections */
+			wait_for_connections();
+
+			/* free all memory we allocated */
+			free_memory();
+
+			if(sigrestart==TRUE){
+
+				/* read the config file */
+				result=read_config_file(config_file);	
+
+				/* exit if there are errors... */
+				if(result==ERROR){
+					syslog(LOG_ERR,"Config file '%s' contained errors, bailing out...",config_file);
+					return STATE_CRITICAL;
+				        }
+			        }
+	
+		        }while(sigrestart==TRUE && sigshutdown==FALSE);
+
+		/* remove pid file */
+		remove_pid_file();
+
+		syslog(LOG_NOTICE,"Daemon shutdown\n");
 	        }
 
 #ifdef HAVE_SSL
@@ -267,10 +343,13 @@ int read_config_file(char *filename){
 	FILE *fp;
 	char config_file[MAX_FILENAME_LENGTH];
 	char input_buffer[MAX_INPUT_BUFFER];
+	char *input_line;
 	char *temp_buffer;
 	char *varname;
 	char *varvalue;
-	int line;
+	int line=0;
+	int len=0;
+	int x=0;
 
 
 	/* open the config file for reading */
@@ -282,21 +361,34 @@ int read_config_file(char *filename){
 		return ERROR;
 	        }
 
-	line=0;
 	while(fgets(input_buffer,MAX_INPUT_BUFFER-1,fp)){
 
 		line++;
+		input_line=input_buffer;
+
+		/* skip leading whitespace */
+		while(isspace(*input_line))
+			++input_line;
+
+		/* trim trailing whitespace */
+		len=strlen(input_line);
+		for(x=len-1;x>=0;x--){
+			if(isspace(input_line[x]))
+				input_line[x]='\x0';
+			else
+				break;
+		        }
 
 		/* skip comments and blank lines */
-		if(input_buffer[0]=='#')
+		if(input_line[0]=='#')
 			continue;
-		if(input_buffer[0]=='\x0')
+		if(input_line[0]=='\x0')
 			continue;
-		if(input_buffer[0]=='\n')
+		if(input_line[0]=='\n')
 			continue;
 
 		/* get the variable name */
-		varname=strtok(input_buffer,"=");
+		varname=strtok(input_line,"=");
 		if(varname==NULL){
 			syslog(LOG_ERR,"No variable name specified in config file '%s' - Line %d\n",filename,line);
 			return ERROR;
@@ -321,15 +413,15 @@ int read_config_file(char *filename){
 
 			/* process the config directory... */
 			if(read_config_dir(config_file)==ERROR)
-				break;
-		        }
+				syslog(LOG_ERR,"Continuing with errors...");
+			}
 
 		/* allow users to specify individual config files to include */
 		else if(!strcmp(varname,"include") || !strcmp(varname,"include_file")){
 
 			/* process the config file... */
 			if(read_config_file(varvalue)==ERROR)
-				break;
+				syslog(LOG_ERR,"Continuing with errors...");
 		        }
 
 		else if(!strcmp(varname,"server_port")){
@@ -339,22 +431,18 @@ int read_config_file(char *filename){
 				return ERROR;
 			        }
 		        }
+		else if(!strcmp(varname,"command_prefix"))
+			command_prefix=strdup(varvalue);
 
 		else if(!strcmp(varname,"server_address")){
                         strncpy(server_address,varvalue,sizeof(server_address) - 1);
                         server_address[sizeof(server_address)-1]='\0';
                         }
 
-		else if(!strcmp(varname,"allowed_hosts")){
-			if(strlen(input_buffer)>sizeof(allowed_hosts)){
-				syslog(LOG_ERR,"Allowed hosts list too long in config file '%s' - Line %d\n",filename,line);
-				return ERROR;
-			        }
-			strncpy(allowed_hosts,varvalue,sizeof(allowed_hosts));
-			allowed_hosts[sizeof(allowed_hosts)-1]='\x0';
-		        }
+                else if(!strcmp(varname,"allowed_hosts"))
+			allowed_hosts=strdup(varvalue);
 
-		else if(strstr(input_buffer,"command[")){
+		else if(strstr(input_line,"command[")){
 			temp_buffer=strtok(varname,"[");
 			temp_buffer=strtok(NULL,"]");
 			if(temp_buffer==NULL){
@@ -389,10 +477,23 @@ int read_config_file(char *filename){
 			        }
 		        }
 
-		else{
-			syslog(LOG_ERR,"Unknown option specified in config file '%s' - Line %d\n",filename,line);
+ 		else if(!strcmp(varname,"connection_timeout")){
+			connection_timeout=atoi(varvalue);
+			if(connection_timeout<1){
+				syslog(LOG_ERR,"Invalid connection_timeout specified in config file '%s' - Line %d\n",filename,line);
+				return ERROR;
+			        }
+		        }
 
-			return ERROR;
+		else if(!strcmp(varname,"allow_weak_random_seed"))
+			allow_weak_random_seed=(atoi(varvalue)==1)?TRUE:FALSE;
+
+		else if(!strcmp(varname,"pid_file"))
+			pid_file=strdup(varvalue);
+
+		else{
+			syslog(LOG_WARNING,"Unknown option specified in config file '%s' - Line %d\n",filename,line);
+			continue;
 		        }
 
 	        }
@@ -410,6 +511,7 @@ int read_config_dir(char *dirname){
 	char config_file[MAX_FILENAME_LENGTH];
 	DIR *dirp;
 	struct dirent *dirfile;
+	struct stat buf;
 	int result=OK;
 	int x;
 
@@ -423,19 +525,18 @@ int read_config_dir(char *dirname){
 	/* process all files in the directory... */
 	while((dirfile=readdir(dirp))!=NULL){
 
+		/* create the full path to the config file or subdirectory */
+		snprintf(config_file,sizeof(config_file)-1,"%s/%s",dirname,dirfile->d_name);
+		config_file[sizeof(config_file)-1]='\x0';
+		stat(config_file, &buf);
+
 		/* process this if it's a config file... */
 		x=strlen(dirfile->d_name);
 		if(x>4 && !strcmp(dirfile->d_name+(x-4),".cfg")){
 
-#ifdef _DIRENT_HAVE_D_TYPE
 			/* only process normal files */
-			if(dirfile->d_type!=DT_REG)
+			if(!S_ISREG(buf.st_mode))
 				continue;
-#endif
-
-			/* create the full path to the config file */
-			snprintf(config_file,sizeof(config_file)-1,"%s/%s",dirname,dirfile->d_name);
-			config_file[sizeof(config_file)-1]='\x0';
 
 			/* process the config file */
 			result=read_config_file(config_file);
@@ -445,18 +546,13 @@ int read_config_dir(char *dirname){
 				break;
 		        }
 
-#ifdef _DIRENT_HAVE_D_TYPE
 		/* recurse into subdirectories... */
-		if(dirfile->d_type==DT_DIR){
+		if(S_ISDIR(buf.st_mode)){
 
 			/* ignore current, parent and hidden directory entries */
 			if(dirfile->d_name[0]=='.')
 				continue;
 
-			/* create the full path to the config directory */
-			snprintf(config_file,sizeof(config_file)-1,"%s/%s",dirname,dirfile->d_name);
-			config_file[sizeof(config_file)-1]='\x0';
-
 			/* process the config directory */
 			result=read_config_dir(config_file);
 
@@ -464,7 +560,6 @@ int read_config_dir(char *dirname){
 			if(result==ERROR)
 				break;
 		        }
-#endif
 		}
 
 	closedir(dirp);
@@ -477,13 +572,25 @@ int read_config_dir(char *dirname){
 int add_command(char *command_name, char *command_line){
 	command *new_command;
 
+	if(command_name==NULL || command_line==NULL)
+		return ERROR;
+
 	/* allocate memory for the new command */
 	new_command=(command *)malloc(sizeof(command));
 	if(new_command==NULL)
 		return ERROR;
 
-	strcpy(new_command->command_name,command_name);
-	strcpy(new_command->command_line,command_line);
+	new_command->command_name=strdup(command_name);
+	if(new_command->command_name==NULL){
+		free(new_command);
+		return ERROR;
+	        }
+	new_command->command_line=strdup(command_line);
+	if(new_command->command_line==NULL){
+		free(new_command->command_name);
+		free(new_command);
+		return ERROR;
+	        }
 
 	/* add new command to head of list in memory */
 	new_command->next=command_list;
@@ -518,9 +625,14 @@ void wait_for_connections(void){
 	int rc;
 	int sock, new_sd;
 	socklen_t addrlen;
-	char connecting_host[16];
 	pid_t pid;
 	int flag=1;
+	fd_set fdread;
+	struct timeval timeout;
+	int retval;
+#ifdef HAVE_LIBWRAP
+	struct request_info req;
+#endif
 
 	/* create a socket for listening */
 	sock=socket(AF_INET,SOCK_STREAM,0);
@@ -531,6 +643,9 @@ void wait_for_connections(void){
 	        exit(STATE_CRITICAL);
 		}
 
+	/* socket should be non-blocking */
+	fcntl(sock,F_SETFL,O_NONBLOCK);
+
         /* set the reuse address flag so we don't get errors when restarting */
         flag=1;
         if(setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&flag,sizeof(flag))<0){
@@ -563,35 +678,72 @@ void wait_for_connections(void){
 	        exit(STATE_CRITICAL);
 		}
 
-	/* log info to syslog facility */
-        syslog(LOG_NOTICE,"Starting up daemon");
-
 	/* log warning about command arguments */
 #ifdef ENABLE_COMMAND_ARGUMENTS
 	if(allow_arguments==TRUE)
 		syslog(LOG_NOTICE,"Warning: Daemon is configured to accept command arguments from clients!");
 #endif
 
-	/* Trap signals */
-	signal(SIGQUIT,sighandler);
-	signal(SIGTERM,sighandler);
+	syslog(LOG_INFO,"Listening for connections on port %d\n",htons(myname.sin_port));
 
-	if(debug==TRUE){
-		syslog(LOG_DEBUG,"Listening for connections on port %d\n",htons(myname.sin_port));
-		syslog(LOG_DEBUG,"Allowing connections from: %s\n",allowed_hosts);
-	        }
+	if(allowed_hosts)
+		syslog(LOG_INFO,"Allowing connections from: %s\n",allowed_hosts);
 
 	/* listen for connection requests - fork() if we get one */
 	while(1){
 
 		/* wait for a connection request */
 	        while(1){
-			new_sd=accept(sock,0,0);
-			if(new_sd>=0 || (errno!=EWOULDBLOCK && errno!=EINTR))
+
+			/* wait until there's something to do */
+			FD_ZERO(&fdread);
+			FD_SET(sock,&fdread);
+			timeout.tv_sec=0;
+			timeout.tv_usec=500000;
+			retval=select(sock+1,&fdread,NULL,&fdread,&timeout);
+
+			/* bail out if necessary */
+			if(sigrestart==TRUE || sigshutdown==TRUE)
 				break;
-			sleep(1);
+
+			/* error */
+			if(retval<0)
+				continue;
+
+			/* accept a new connection request */
+			new_sd=accept(sock,0,0);
+
+			/* some kind of error occurred... */
+			if(new_sd<0){
+
+				/* bail out if necessary */
+				if(sigrestart==TRUE || sigshutdown==TRUE)
+					break;
+
+				/* retry */
+				if(errno==EWOULDBLOCK || errno==EINTR)
+					continue;
+
+				/* socket is nonblocking and we don't have a connection yet */
+				if(errno==EAGAIN)
+					continue;
+
+				/* fix for HP-UX 11.0 - just retry */
+				if(errno==ENOBUFS)
+					continue;
+
+				/* else handle the error later */
+				break;
+			        }
+
+			/* connection was good */
+			break;
 		        }
 
+		/* bail out if necessary */
+		if(sigrestart==TRUE || sigshutdown==TRUE)
+			break;
+
 		/* child process should handle the connection */
     		pid=fork();
     		if(pid==0){
@@ -612,6 +764,11 @@ void wait_for_connections(void){
 					return;
 				        }
 
+				/* handle signals */
+				signal(SIGQUIT,child_sighandler);
+				signal(SIGTERM,child_sighandler);
+				signal(SIGHUP,child_sighandler);
+
 				/* grandchild does not need to listen for connections, so close the socket */
 				close(sock);  
 
@@ -636,33 +793,62 @@ void wait_for_connections(void){
 				if(debug==TRUE)
 					syslog(LOG_DEBUG,"Connection from %s port %d",inet_ntoa(nptr->sin_addr),nptr->sin_port);
 
-				/* is this is a blessed machine? */
-				snprintf(connecting_host,sizeof(connecting_host),"%s",inet_ntoa(nptr->sin_addr));
-				connecting_host[sizeof(connecting_host)-1]='\x0';
+                                /* is this is a blessed machine? */
+				if(allowed_hosts){
 
-				if(!is_an_allowed_host(connecting_host)){
+					if(!is_an_allowed_host(inet_ntoa(nptr->sin_addr))){
 
-				        /* log error to syslog facility */
-					syslog(LOG_ERR,"Host %s is not allowed to talk to us!",connecting_host);
-			                }
-				else{
+                                               /* log error to syslog facility */
+                                               syslog(LOG_ERR,"Host %s is not allowed to talk to us!",inet_ntoa(nptr->sin_addr));
 
-				        /* log info to syslog facility */
-					if(debug==TRUE)
-						syslog(LOG_DEBUG,"Host address checks out ok");
+                                               /* log info to syslog facility */
+					       if(debug==TRUE)
+						       syslog(LOG_DEBUG,"Connection from %s closed.",inet_ntoa(nptr->sin_addr));
 
-				        /* handle the client connection */
-					handle_connection(new_sd);
-			                }
+					       /* close socket prior to exiting */
+                                               close(new_sd);
+
+					       exit(STATE_OK);
+				               }
+                                       else{
+
+                                               /* log info to syslog facility */
+                                               if(debug==TRUE)
+                                                       syslog(LOG_DEBUG,"Host address is in allowed_hosts");
+				               }
+				        }
+
+#ifdef HAVE_LIBWRAP
+
+				/* Check whether or not connections are allowed from this host */
+				request_init(&req,RQ_DAEMON,"nrpe",RQ_FILE,new_sd,0);
+				fromhost(&req);
+
+				if(!hosts_access(&req)){
+
+					syslog(LOG_DEBUG,"Connection refused by TCP wrapper");
+
+					/* refuse the connection */
+					refuse(&req);
+					close(new_sd);
+
+					/* should not be reached */
+					syslog(LOG_ERR,"libwrap refuse() returns!");
+					exit(STATE_CRITICAL);
+					}
+#endif
+
+				/* handle the client connection */
+				handle_connection(new_sd);
 
 				/* log info to syslog facility */
 				if(debug==TRUE)
-					syslog(LOG_DEBUG,"Connection from %s closed.",connecting_host);
+					syslog(LOG_DEBUG,"Connection from %s closed.",inet_ntoa(nptr->sin_addr));
 
 				/* close socket prior to exiting */
 				close(new_sd);
 
-				return;
+				exit(STATE_OK);
     			        }
 
 			/* first child returns immediately, grandchild is inherited by INIT process -> no zombies... */
@@ -680,14 +866,80 @@ void wait_for_connections(void){
 		        }
   		}
 
-	/* we shouldn't ever get here... */
-	syslog(LOG_NOTICE,"Terminating");
+	/* close the socket we're listening on */
+	close(sock);
 
 	return;
 	}
 
 
 
+/* checks to see if a given host is allowed to talk to us */
+int is_an_allowed_host(char *connecting_host){
+	char *temp_buffer=NULL;
+	char *temp_ptr=NULL;
+	int result=0;
+        struct hostent *myhost;
+	char **pptr=NULL;
+	char *save_connecting_host=NULL;
+	struct in_addr addr;
+	
+	/* make sure we have something */
+	if(connecting_host==NULL)
+		return 0;
+	if(allowed_hosts==NULL)
+		return 1;
+
+	if((temp_buffer=strdup(allowed_hosts))==NULL)
+		return 0;
+	
+	/* try and match IP addresses first */
+	for(temp_ptr=strtok(temp_buffer,",");temp_ptr!=NULL;temp_ptr=strtok(NULL,",")){
+
+		if(!strcmp(connecting_host,temp_ptr)){
+			result=1;
+			break;
+		        }
+	        }
+
+	/* try DNS lookups if needed */
+	if(result==0){
+
+		free(temp_buffer);
+		if((temp_buffer=strdup(allowed_hosts))==NULL)
+			return 0;
+
+		save_connecting_host=strdup(connecting_host);
+		for(temp_ptr=strtok(temp_buffer,",");temp_ptr!=NULL;temp_ptr=strtok(NULL,",")){
+
+			myhost=gethostbyname(temp_ptr);
+			if(myhost!=NULL){
+
+				/* check all addresses for the host... */
+				for(pptr=myhost->h_addr_list;*pptr!=NULL;pptr++){
+					memcpy(&addr, *pptr, sizeof(addr));
+					if(!strcmp(save_connecting_host, inet_ntoa(addr))){
+						result=1;
+						break;
+					        }
+				        }
+			        }
+
+			if(result==1)
+				break;
+		        }
+
+		strcpy(connecting_host, save_connecting_host);
+		free(save_connecting_host);
+	        }
+
+	free(temp_buffer);
+
+	return result;
+        }
+
+
+
 /* handles a client connection */
 void handle_connection(int sock){
         u_int32_t calculated_crc32;
@@ -720,12 +972,20 @@ void handle_connection(int sock){
 	fcntl(sock,F_SETFL,O_NONBLOCK);
 #endif
 
+	/* set connection handler */
+	signal(SIGALRM,my_connection_sighandler);
+	alarm(connection_timeout);
+
 #ifdef HAVE_SSL
 	/* do SSL handshake */
 	if(result==STATE_OK && use_ssl==TRUE){
 		if((ssl=SSL_new(ctx))!=NULL){
 			SSL_set_fd(ssl,sock);
-			if((rc=SSL_accept(ssl))!=1){
+
+			/* keep attempting the request if needed */
+                        while(((rc=SSL_accept(ssl))!=1) && (SSL_get_error(ssl,rc)==SSL_ERROR_WANT_READ));
+
+			if(rc!=1){
 				syslog(LOG_ERR,"Error: Could not complete SSL handshake. %d\n",SSL_get_error(ssl,rc));
 #ifdef DEBUG
 				errfp=fopen("/tmp/err.log","w");
@@ -751,8 +1011,9 @@ void handle_connection(int sock){
 	if(use_ssl==FALSE)
 		rc=recvall(sock,(char *)&receive_packet,&bytes_to_recv,socket_timeout);
 #ifdef HAVE_SSL
-	else
-		rc=SSL_read(ssl,&receive_packet,bytes_to_recv);
+	else{
+                while(((rc=SSL_read(ssl,&receive_packet,bytes_to_recv))<=0) && (SSL_get_error(ssl,rc)==SSL_ERROR_WANT_READ));
+		}
 #endif
 
 	/* recv() error or client disconnect */
@@ -762,9 +1023,11 @@ void handle_connection(int sock){
 		syslog(LOG_ERR,"Could not read request from client, bailing out...");
 
 #ifdef HAVE_SSL
-		SSL_shutdown(ssl);
-		SSL_free(ssl);
-		syslog(LOG_INFO,"INFO: SSL Socket Shutdown.\n");
+		if(ssl){
+			SSL_shutdown(ssl);
+			SSL_free(ssl);
+			syslog(LOG_INFO,"INFO: SSL Socket Shutdown.\n");
+			}
 #endif
 
 		return;
@@ -777,8 +1040,10 @@ void handle_connection(int sock){
 		syslog(LOG_ERR,"Data packet from client was too short, bailing out...");
 
 #ifdef HAVE_SSL
-		SSL_shutdown(ssl);
-		SSL_free(ssl);
+		if(ssl){
+			SSL_shutdown(ssl);
+			SSL_free(ssl);
+			}
 #endif
 
 		return;
@@ -807,8 +1072,10 @@ void handle_connection(int sock){
 	                }
 
 #ifdef HAVE_SSL
-		SSL_shutdown(ssl);
-		SSL_free(ssl);
+		if(ssl){
+			SSL_shutdown(ssl);
+			SSL_free(ssl);
+			}
 #endif
 
 		return;
@@ -818,6 +1085,9 @@ void handle_connection(int sock){
 	if(debug==TRUE)
 		syslog(LOG_DEBUG,"Host is asking for command '%s' to be run...",receive_packet.buffer);
 
+	/* disable connection alarm - a new alarm will be setup during my_system */
+	alarm(0);
+
 	/* if this is the version check command, just spew it out */
 	if(!strcmp(command_name,NRPE_HELLO_COMMAND)){
 
@@ -849,7 +1119,10 @@ void handle_connection(int sock){
 		else{
 
 			/* process command line */
-			strncpy(raw_command,temp_command->command_line,sizeof(raw_command)-1);
+			if(command_prefix==NULL)
+				strncpy(raw_command,temp_command->command_line,sizeof(raw_command)-1);
+			else
+				snprintf(raw_command,sizeof(raw_command)-1,"%s %s",command_prefix,temp_command->command_line);
 			raw_command[sizeof(raw_command)-1]='\x0';
 			process_macros(raw_command,processed_command,sizeof(processed_command));
 
@@ -928,8 +1201,10 @@ void handle_connection(int sock){
 #endif
 
 #ifdef HAVE_SSL
-	SSL_shutdown(ssl);
-	SSL_free(ssl);
+	if(ssl){
+		SSL_shutdown(ssl);
+		SSL_free(ssl);
+		}
 #endif
 
 	/* log info to syslog facility */
@@ -941,39 +1216,6 @@ void handle_connection(int sock){
 
 
 
-/* checks to see if a given host is allowed to talk to us */
-int is_an_allowed_host(char *connecting_host){
-	char temp_buffer[MAX_INPUT_BUFFER];
-	char *temp_ptr;
-
-	strncpy(temp_buffer,allowed_hosts,sizeof(temp_buffer));
-	temp_buffer[sizeof(temp_buffer)-1]='\x0';
-
-	for(temp_ptr=strtok(temp_buffer,",");temp_ptr!=NULL;temp_ptr=strtok(NULL,",")){
-		if(!strcmp(connecting_host,temp_ptr))
-			return 1;
-	        }
-
-	return 0;
-        }
-
-
-
-/* handle signals */
-void sighandler(int sig){
-
-	/* free all memory we allocated */
-	free_memory();
-	
-	/* terminate... */
-	exit(0);
-
-	/* so the compiler doesn't complain.. */
-	return;
-        }
-
-
-
 /* free all allocated memory */
 void free_memory(void){
 	command *this_command;
@@ -983,16 +1225,21 @@ void free_memory(void){
 	this_command=command_list;
 	while(this_command!=NULL){
 		next_command=this_command->next;
+		if(this_command->command_name)
+			free(this_command->command_name);
+		if(this_command->command_line)
+			free(this_command->command_line);
 		free(this_command);
 		this_command=next_command;
 	        }
 
+	command_list=NULL;
+
 	return;
         }
 
 
 
-
 /* executes a system command via popen(), but protects against timeouts */
 int my_system(char *command,int timeout,int *early_timeout,char *output,int output_length){
         pid_t pid;
@@ -1000,6 +1247,7 @@ int my_system(char *command,int timeout,int *early_timeout,char *output,int outp
 	int result;
 	extern int errno;
 	char buffer[MAX_INPUT_BUFFER];
+	char temp_buffer[MAX_INPUT_BUFFER];
 	int fd[2];
 	FILE *fp;
 	int bytes_read=0;
@@ -1074,9 +1322,12 @@ int my_system(char *command,int timeout,int *early_timeout,char *output,int outp
 		        }
 		else{
 
-			/* read in the first line of output from the command */
-			strcpy(buffer,"");
-			fgets(buffer,sizeof(buffer)-1,fp);
+			/* read all lines of output - supports Nagios 3.x multiline output */
+			while((bytes_read=fread(buffer,1,sizeof(buffer)-1,fp))>0){
+
+				/* write the output back to the parent process */
+				write(fd[1],buffer,bytes_read);
+				}
 
 			/* close the command and get termination status */
 			status=pclose(fp);
@@ -1086,9 +1337,6 @@ int my_system(char *command,int timeout,int *early_timeout,char *output,int outp
 				result=STATE_CRITICAL;
 			else
 				result=WEXITSTATUS(status);
-
-			/* write the output back to the parent process */
-			write(fd[1],buffer,strlen(buffer)+1);
 		        }
 
 		/* close pipe for writing */
@@ -1097,6 +1345,7 @@ int my_system(char *command,int timeout,int *early_timeout,char *output,int outp
 		/* reset the alarm */
 		alarm(0);
 
+		/* return plugin exit code to parent process */
 		exit(result);
 	        }
 
@@ -1164,6 +1413,14 @@ void my_system_sighandler(int sig){
         }
 
 
+/* handle errors where connection takes too long */
+void my_connection_sighandler(int sig) {
+
+	syslog(LOG_ERR,"Connection has taken too long to establish. Exiting...");
+
+	exit(STATE_CRITICAL);
+	}
+
 
 /* drops privileges */
 int drop_privileges(char *user, char *group){
@@ -1182,6 +1439,7 @@ int drop_privileges(char *user, char *group){
 				gid=(gid_t)(grp->gr_gid);
 			else
 				syslog(LOG_ERR,"Warning: Could not get group entry for '%s'",group);
+			endgrent();
 		        }
 
 		/* else we were passed the GID */
@@ -1207,16 +1465,17 @@ int drop_privileges(char *user, char *group){
 				uid=(uid_t)(pw->pw_uid);
 			else
 				syslog(LOG_ERR,"Warning: Could not get passwd entry for '%s'",user);
+			endpwent();
 		        }
 
 		/* else we were passed the UID */
 		else
 			uid=(uid_t)atoi(user);
 			
-#ifdef HAVE_INITGROUPS
-
+		/* set effective user ID if other than current EUID */
 		if(uid!=geteuid()){
 
+#ifdef HAVE_INITGROUPS
 			/* initialize supplementary groups */
 			if(initgroups(user,gid)==-1){
 				if(errno==EPERM)
@@ -1226,17 +1485,88 @@ int drop_privileges(char *user, char *group){
 					return ERROR;
 			                }
 	                        }
-		        }
 #endif
 
-		if(setuid(uid)==-1)
-			syslog(LOG_ERR,"Warning: Could not set effective UID=%d",(int)uid);
+			if(setuid(uid)==-1)
+				syslog(LOG_ERR,"Warning: Could not set effective UID=%d",(int)uid);
+		        }
 	        }
 
 	return OK;
         }
 
 
+/* write an optional pid file */
+int write_pid_file(void){
+	int fd;
+	int result=0;
+	pid_t pid=0;
+	char pbuf[16];
+
+	/* no pid file was specified */
+	if(pid_file==NULL)
+		return OK;
+
+	/* read existing pid file */
+	if((fd=open(pid_file,O_RDONLY))>=0){
+
+		result=read(fd,pbuf,(sizeof pbuf)-1);
+
+		close(fd);
+
+		if(result>0){
+
+			pbuf[result]='\x0';
+			pid=(pid_t)atoi(pbuf);
+
+			/* if previous process is no longer running running, remove the old pid file */
+			if(pid && (pid==getpid() || kill(pid,0)<0))
+				unlink(pid_file);
+
+			/* previous process is still running */
+			else{
+				syslog(LOG_ERR,"There's already an NRPE server running (PID %lu).  Bailing out...",(unsigned long)pid);
+				return ERROR;
+			        }
+		        }
+	        } 
+
+	/* write new pid file */
+	if((fd=open(pid_file,O_WRONLY | O_CREAT,0644))>=0){
+		sprintf(pbuf,"%d\n",(int)getpid());
+		write(fd,pbuf,strlen(pbuf));
+		close(fd);
+		wrote_pid_file=TRUE;
+	        }
+	else{
+		syslog(LOG_ERR,"Cannot write to pidfile '%s' - check your privileges.",pid_file);
+	        }
+
+	return OK;
+        }
+
+
+
+/* remove pid file */
+int remove_pid_file(void){
+
+	/* no pid file was specified */
+	if(pid_file==NULL)
+		return OK;
+
+	/* pid file was not written */
+	if(wrote_pid_file==FALSE)
+		return OK;
+
+	/* remove existing pid file */
+	if(unlink(pid_file)==-1){
+		syslog(LOG_ERR,"Cannot remove pidfile '%s' - check your privileges.",pid_file);
+		return ERROR;
+	        }
+
+	return OK;
+        }
+
 
 
 /* bail if daemon is running as root */
@@ -1257,6 +1587,59 @@ int check_privileges(void){
 
 
 
+/* handle signals (parent process) */
+void sighandler(int sig){
+	static char *sigs[]={"EXIT","HUP","INT","QUIT","ILL","TRAP","ABRT","BUS","FPE","KILL","USR1","SEGV","USR2","PIPE","ALRM","TERM","STKFLT","CHLD","CONT","STOP","TSTP","TTIN","TTOU","URG","XCPU","XFSZ","VTALRM","PROF","WINCH","IO","PWR","UNUSED","ZERR","DEBUG",(char *)NULL};
+	int i;
+	char temp_buffer[MAX_INPUT_BUFFER];
+
+	if(sig<0)
+		sig=-sig;
+
+	for(i=0;sigs[i]!=(char *)NULL;i++);
+
+	sig%=i;
+
+	/* we received a SIGHUP, so restart... */
+	if(sig==SIGHUP){
+
+		sigrestart=TRUE;
+
+		syslog(LOG_NOTICE,"Caught SIGHUP - restarting...\n");
+	        }
+
+	/* else begin shutting down... */
+	if(sig==SIGTERM){
+
+		/* if shutdown is already true, we're in a signal trap loop! */
+		if(sigshutdown==TRUE)
+			exit(STATE_CRITICAL);
+
+		sigshutdown=TRUE;
+
+		syslog(LOG_NOTICE,"Caught SIG%s - shutting down...\n",sigs[sig]);
+	        }
+
+	return;
+        }
+
+
+
+/* handle signals (child processes) */
+void child_sighandler(int sig){
+
+	/* free all memory we allocated */
+	free_memory();
+
+	/* terminate */
+	exit(0);
+	
+	/* so the compiler doesn't complain... */
+	return;
+        }
+
+
+
 /* tests whether or not a client request is valid */
 int validate_request(packet *pkt){
         u_int32_t packet_crc32;
@@ -1437,7 +1820,7 @@ int process_arguments(int argc, char **argv){
 	int c=1;
 	int have_mode=FALSE;
 
-#ifdef HAVE_GETOPT_H
+#ifdef HAVE_GETOPT_LONG
 	int option_index=0;
 	static struct option long_options[]={
 		{"config", required_argument, 0, 'c'},
@@ -1454,10 +1837,10 @@ int process_arguments(int argc, char **argv){
 	if(argc<2)
 		return ERROR;
 
-	snprintf(optchars,MAX_INPUT_BUFFER,"c:nidhl");
+	snprintf(optchars,MAX_INPUT_BUFFER,"c:hVldin");
 
 	while(1){
-#ifdef HAVE_GETOPT_H
+#ifdef HAVE_GETOPT_LONG
 		c=getopt_long(argc,argv,optchars,long_options,&option_index);
 #else
 		c=getopt(argc,argv,optchars);
